@@ -13,8 +13,6 @@ let ComposerPageState = function() {
     this.init.apply(this, arguments);
 };
 ComposerPageState.KEYWORD_SPLIT_REGEX = /[\s]+/g;
-ComposerPageState.QUOTE_START = "\x91";  // private use one
-ComposerPageState.QUOTE_END = "\x92";    // private use two
 ComposerPageState.QUOTE_MARKER = "\x7f"; // delete
 ComposerPageState.PROTOCOL_REGEX = /^(aim|apt|bitcoin|cvs|ed2k|ftp|file|finger|git|gtalk|http|https|irc|ircs|irc6|lastfm|ldap|ldaps|magnet|news|nntp|rsync|sftp|skype|smb|sms|svn|telnet|tftp|ssh|webcal|xmpp):/i;
 // Taken from Geary.HTML.URL_REGEX, without the inline modifier (?x)
@@ -30,8 +28,6 @@ ComposerPageState.prototype = {
         this.quotePart = null;
         this.focusedPart = null;
 
-        this.undoEnabled = false;
-        this.redoEnabled = false;
         this.selections = new Map();
         this.nextSelectionId = 0;
         this.cursorContext = null;
@@ -43,17 +39,6 @@ ComposerPageState.prototype = {
                 e.preventDefault();
             }
         }, true);
-
-        let modifiedId = null;
-        this.bodyObserver = new MutationObserver(function() {
-            if (modifiedId == null) {
-                modifiedId = window.setTimeout(function() {
-                    state.documentModified();
-                    state.checkCommandStack();
-                    modifiedId = null;
-                }, 1000);
-            }
-        });
     },
     loaded: function() {
         let state = this;
@@ -138,8 +123,16 @@ ComposerPageState.prototype = {
             cursor.parentNode.removeChild(cursor);
         }
 
-        // Enable editing and observation machinery only after
-        // modifying the body above.
+
+        // Enable editing only after modifying the body above.
+        this.setEditable(true);
+
+        PageState.prototype.loaded.apply(this, []);
+    },
+    setEditable: function(enabled) {
+        if (!enabled) {
+            this.stopBodyObserver();
+        }
         this.bodyPart.contentEditable = true;
         if (this.signaturePart != null) {
             this.signaturePart.contentEditable = true;
@@ -147,16 +140,11 @@ ComposerPageState.prototype = {
         if (this.quotePart != null) {
             this.quotePart.contentEditable = true;
         }
-        let config = {
-            attributes: true,
-            childList: true,
-            characterData: true,
-            subtree: true
-        };
-        this.bodyObserver.observe(document.body, config);
-
-        // Chain up
-        PageState.prototype.loaded.apply(this, []);
+        if (enabled) {
+            // Enable modification observation only after the document
+            // has been set editable as WebKit will alter some attrs
+            this.startBodyObserver();
+        }
     },
     undo: function() {
         document.execCommand("undo", false, null);
@@ -212,6 +200,12 @@ ComposerPageState.prototype = {
             element.setAttribute("type", "cite");
         }
     },
+    insertOrderedList: function() {
+        document.execCommand("insertOrderedList", false, null);
+    },
+    insertUnorderedList: function() {
+        document.execCommand("insertUnorderedList", false, null);
+    },
     updateSignature: function(signature) {
         if (this.signaturePart != null) {
             console.log(signature);
@@ -299,6 +293,10 @@ ComposerPageState.prototype = {
         }
     },
     cleanContent: function() {
+        // Prevent any modification signals being sent when mutating
+        // the document below.
+        this.stopBodyObserver();
+
         ComposerPageState.cleanPart(this.bodyPart, false);
         ComposerPageState.linkify(this.bodyPart);
 
@@ -330,7 +328,8 @@ ComposerPageState.prototype = {
         return parent.innerHTML;
     },
     getText: function() {
-        return ComposerPageState.htmlToQuotedText(document.body);
+        let text = ComposerPageState.htmlToText(document.body);
+        return ComposerPageState.replaceNonBreakingSpace(text);
     },
     setRichText: function(enabled) {
         if (enabled) {
@@ -339,21 +338,6 @@ ComposerPageState.prototype = {
             document.body.classList.add("plain");
         }
     },
-    checkCommandStack: function() {
-        let canUndo = document.queryCommandEnabled("undo");
-        let canRedo = document.queryCommandEnabled("redo");
-
-        if (canUndo != this.undoEnabled || canRedo != this.redoEnabled) {
-            this.undoEnabled = canUndo;
-            this.redoEnabled = canRedo;
-            window.webkit.messageHandlers.commandStackChanged.postMessage(
-                this.undoEnabled + "," + this.redoEnabled
-            );
-        }
-    },
-    documentModified: function(element) {
-        window.webkit.messageHandlers.documentModified.postMessage(null);
-    },
     selectionChanged: function() {
         PageState.prototype.selectionChanged.apply(this, []);
 
@@ -456,66 +440,122 @@ ComposerPageState.cleanPart = function(part, removeIfEmpty) {
 };
 
 /**
- * Convert a HTML DOM tree to plain text with delineated quotes.
+ * Gets plain text that adequately represents the information in the HTML
  *
- * Lines are delinated using LF. Quoted lines are prefixed with
- * `ComposerPageState.QUOTE_MARKER`, where the number of markers
- * indicates the depth of nesting of the quote.
+ * Asterisks are inserted around bold text, slashes around italic text, and
+ * underscores around underlined text. Link URLs are inserted after the link
+ * text.
  *
- * This will modify/reset the DOM, since it ultimately requires
- * stuffing `QUOTE_MARKER` into existing paragraphs and getting it
- * back out in a way that preserves the visual presentation.
+ * Each line of a blockquote is prefixed with
+ * `ComposerPageState.QUOTE_MARKER`, where the number of markers indicates
+ * the depth of nesting of the quote.
  */
-ComposerPageState.htmlToQuotedText = function(root) {
-    // XXX It would be nice to just clone the root and modify that, or
-    // see if we can implement this some other way so as to not modify
-    // the DOM at all, but currently unit test show that the results
-    // are not the same if we work on a clone, likely because of the
-    // use of HTMLElement::innerText. Need to look into it more.
+ComposerPageState.htmlToText = function(root) {
+    let parentStyle = window.getComputedStyle(root);
+    let text = "";
 
-    let savedDoc = root.innerHTML;
-    let blockquotes = root.querySelectorAll("blockquote");
-    let nbq = blockquotes.length;
-    let bqtexts = new Array(nbq);
-
-    // Get text of blockquotes and pull them out of DOM.  They are
-    // replaced with tokens deliminated with the characters
-    // QUOTE_START and QUOTE_END (from a unicode private use block).
-    // We need to get the text while they're still in the DOM to get
-    // newlines at appropriate places.  We go through the list of
-    // blockquotes from the end so that we get the innermost ones
-    // first.
-    for (let i = nbq - 1; i >= 0; i--) {
-        let bq = blockquotes.item(i);
-        let text = bq.innerText;
-        if (text.substr(-1, 1) == "\n") {
-            text = text.slice(0, -1);
-        } else {
-            console.debug(
-                "  no newline at end of quote: " +
-                    text.length > 0
-                    ? "0x" + text.codePointAt(text.length - 1).toString(16)
-                    : "empty line"
-            );
-        }
-        bqtexts[i] = text;
-
-        bq.innerText = (
-            ComposerPageState.QUOTE_START
-                + i.toString()
-                + ComposerPageState.QUOTE_END
+    for (let node of (root.childNodes || [])) {
+        let isBlock = (
+            node instanceof Element
+                && window.getComputedStyle(node).display == "block"
+                && node.innerText
         );
+        if (isBlock) {
+            // Make sure there's a newline before the element
+            if (text != "" && text.substr(-1) != "\n") {
+                text += "\n";
+            }
+        }
+        switch (node.nodeName.toLowerCase()) {
+            case "#text":
+                let nodeText = node.nodeValue;
+                switch (parentStyle.whiteSpace) {
+                    case 'normal':
+                    case 'nowrap':
+                    case 'pre-line':
+                        // Only space, tab, carriage return, and newline collapse
+                        // https://www.w3.org/TR/2011/REC-CSS2-20110607/text.html#white-space-model
+                        nodeText = nodeText.replace(/[ \t\r\n]+/g, " ");
+                        if (nodeText == " " && " \t\r\n".includes(text.substr(-1)))
+                            break; // There's already whitespace here
+                        if (node == root.firstChild)
+                            nodeText = nodeText.replace(/^ /, "");
+                        if (node == root.lastChild)
+                            nodeText = nodeText.replace(/ $/, "");
+                        // Fall through
+                    default:
+                        text += nodeText;
+                        break;
+                }
+                break;
+            case "a":
+                if (node.closest("body.plain")) {
+                    text += ComposerPageState.htmlToText(node);
+                } else if (node.textContent == node.href) {
+                    text += "<" + node.href + ">";
+                } else {
+                    text += ComposerPageState.htmlToText(node);
+                    text += " <" + node.href + ">";
+                }
+                break;
+            case "b":
+            case "strong":
+                if (node.closest("body.plain")) {
+                    text += ComposerPageState.htmlToText(node);
+                } else {
+                    text += "*" + ComposerPageState.htmlToText(node) + "*";
+                }
+                break;
+            case "blockquote":
+                let bqText = ComposerPageState.htmlToText(node);
+                // If there is a newline at the end of the quote, remove it
+                // After this switch we ensure that there is a newline after the quote
+                bqText = bqText.replace(/\n$/, "");
+                let lines = bqText.split("\n");
+                for (let i = 0; i < lines.length; i++)
+                    lines[i] = ComposerPageState.QUOTE_MARKER + lines[i];
+                text += lines.join("\n");
+                break;
+            case "br":
+                text += "\n";
+                break;
+            case "i":
+            case "em":
+                if (node.closest("body.plain")) {
+                    text += ComposerPageState.htmlToText(node);
+                } else {
+                    text += "/" + ComposerPageState.htmlToText(node) + "/";
+                }
+                break;
+            case "u":
+                if (node.closest("body.plain")) {
+                    text += ComposerPageState.htmlToText(node);
+                } else {
+                    text += "_" + ComposerPageState.htmlToText(node) + "_";
+                }
+                break;
+            case "#comment":
+                break;
+            default:
+                text += ComposerPageState.htmlToText(node);
+                break;
+        }
+        if (isBlock) {
+            // Ensure that the last character is a newline
+            if (text.substr(-1) != "\n") {
+                text += "\n";
+            }
+            if (node.nodeName.toLowerCase() == "p") {
+                // Ensure that the last two characters are newlines
+                if (text.substr(-2, 1) != "\n") {
+                    text += "\n";
+                }
+            }
+        }
     }
 
-    // Reassemble plain text out of parts, and replace non-breaking
-    // space with regular space.
-    let text = ComposerPageState.resolveNesting(root.innerText, bqtexts);
-
-    // Reassemble DOM now we have the plain text
-    root.innerHTML = savedDoc;
-
-    return ComposerPageState.replaceNonBreakingSpace(text);
-};
+    return text;
+}
 
 // Linkifies "plain text" link
 ComposerPageState.linkify = function(node) {
@@ -565,46 +605,6 @@ ComposerPageState.linkify = function(node) {
     }
 };
 
-ComposerPageState.resolveNesting = function(text, values) {
-    let tokenregex = new RegExp(
-        "(.?)" +
-            ComposerPageState.QUOTE_START +
-            "([0-9]*)" +
-            ComposerPageState.QUOTE_END +
-            "(?=(.?))", "g"
-    );
-    return text.replace(tokenregex, function(match, p1, p2, p3, offset, str) {
-        let key = new Number(p2);
-        let prevChars = p1;
-        let nextChars = p3;
-        let insertNext = "";
-        // Make sure there's a newline before and after the quote.
-        if (prevChars != "" && prevChars != "\n")
-            prevChars = prevChars + "\n";
-        if (nextChars != "" && nextChars != "\n")
-            insertNext = "\n";
-
-        let value = "";
-        if (key >= 0 && key < values.length) {
-            let nested = ComposerPageState.resolveNesting(values[key], values);
-            value = prevChars + ComposerPageState.quoteLines(nested) + insertNext;
-        } else {
-            console.error("Regex error in denesting blockquotes: Invalid key");
-        }
-        return value;
-    });
-};
-
-/**
- * Prefixes each NL-delineated line with `ComposerPageState.QUOTE_MARKER`.
- */
-ComposerPageState.quoteLines = function(text) {
-    let lines = text.split("\n");
-    for (let i = 0; i < lines.length; i++)
-        lines[i] = ComposerPageState.QUOTE_MARKER + lines[i];
-    return lines.join("\n");
-};
-
 /**
  * Converts all non-breaking space chars to plain spaces.
  */
